@@ -54,7 +54,7 @@
 #endif
 
 
-static int init_geometry_boot(struct bootsector *boot, struct device *dev,
+static int init_geometry_boot(union bootsector *boot, struct device *dev,
 			       int sectors0, int rate_0, int rate_any,
 			       unsigned long *tot_sectors, int keepBoot)
 {
@@ -65,45 +65,47 @@ static int init_geometry_boot(struct bootsector *boot, struct device *dev,
 	int j;
 	int sum;
 
-	set_word(boot->nsect, dev->sectors);
-	set_word(boot->nheads, dev->heads);
+	set_word(boot->boot.nsect, dev->sectors);
+	set_word(boot->boot.nheads, dev->heads);
 
-	*tot_sectors = dev->heads * dev->sectors * dev->tracks - DWORD(nhs);
+#ifdef HAVE_ASSERT_H
+	assert(*tot_sectors != 0);
+#endif
 
 	if (*tot_sectors < 0x10000){
-		set_word(boot->psect, *tot_sectors);
-		set_dword(boot->bigsect, 0);
+		set_word(boot->boot.psect, *tot_sectors);
+		set_dword(boot->boot.bigsect, 0);
 	} else {
-		set_word(boot->psect, 0);
-		set_dword(boot->bigsect, *tot_sectors);
+		set_word(boot->boot.psect, 0);
+		set_dword(boot->boot.bigsect, *tot_sectors);
 	}
 
 	if (dev->use_2m & 0x7f){
 		int bootOffset;
-		strncpy(boot->banner, "2M-STV04", 8);
-		boot->ext.old.res_2m = 0;
-		boot->ext.old.fmt_2mf = 6;
+		strncpy(boot->boot.banner, "2M-STV04", 8);
+		boot->boot.ext.old.res_2m = 0;
+		boot->boot.ext.old.fmt_2mf = 6;
 		if ( dev->sectors % ( ((1 << dev->ssize) + 3) >> 2 ))
-			boot->ext.old.wt = 1;
+			boot->boot.ext.old.wt = 1;
 		else
-			boot->ext.old.wt = 0;
-		boot->ext.old.rate_0= rate_0;
-		boot->ext.old.rate_any= rate_any;
-		if (boot->ext.old.rate_any== 2 )
-			boot->ext.old.rate_any= 1;
+			boot->boot.ext.old.wt = 0;
+		boot->boot.ext.old.rate_0= rate_0;
+		boot->boot.ext.old.rate_any= rate_any;
+		if (boot->boot.ext.old.rate_any== 2 )
+			boot->boot.ext.old.rate_any= 1;
 		i=76;
 
 		/* Infp0 */
-		set_word(boot->ext.old.Infp0, i);
-		uchr(boot)[i++] = sectors0;
-		uchr(boot)[i++] = 108;
+		set_word(boot->boot.ext.old.Infp0, i);
+		boot->bytes[i++] = sectors0;
+		boot->bytes[i++] = 108;
 		for(j=1; j<= sectors0; j++)
-			uchr(boot)[i++] = j;
+			boot->bytes[i++] = j;
 
-		set_word(boot->ext.old.InfpX, i);
-		
-		uchr(boot)[i++] = 64;
-		uchr(boot)[i++] = 3;
+		set_word(boot->boot.ext.old.InfpX, i);
+
+		boot->bytes[i++] = 64;
+		boot->bytes[i++] = 3;
 		nb_renum = i++;
 		sector2 = dev->sectors;
 		size2 = dev->ssize;
@@ -111,38 +113,38 @@ static int init_geometry_boot(struct bootsector *boot, struct device *dev,
 		while( sector2 ){
 			while ( sector2 < (1 << size2) >> 2 )
 				size2--;
-			uchr(boot)[i++] = 128 + j;
-			uchr(boot)[i++] = j++;
-			uchr(boot)[i++] = size2;
+			boot->bytes[i++] = 128 + j;
+			boot->bytes[i++] = j++;
+			boot->bytes[i++] = size2;
 			sector2 -= (1 << size2) >> 2;
 		}
-		uchr(boot)[nb_renum] = ( i - nb_renum - 1 ) / 3;
+		boot->bytes[nb_renum] = ( i - nb_renum - 1 ) / 3;
 
-		set_word(boot->ext.old.InfTm, i);
+		set_word(boot->boot.ext.old.InfTm, i);
 
 		sector2 = dev->sectors;
 		size2= dev->ssize;
 		while(sector2){
 			while ( sector2 < 1 << ( size2 - 2) )
 				size2--;
-			uchr(boot)[i++] = size2;
+			boot->bytes[i++] = size2;
 			sector2 -= 1 << (size2 - 2 );
 		}
-		
-		set_word(boot->ext.old.BootP,i);
+
+		set_word(boot->boot.ext.old.BootP,i);
 		bootOffset = i;
 
-		/* checksum */		
-		for (sum=0, j=64; j<i; j++) 
-			sum += uchr(boot)[j];/* checksum */
-		boot->ext.old.CheckSum=-sum;
+		/* checksum */
+		for (sum=0, j=64; j<i; j++)
+			sum += boot->bytes[j];/* checksum */
+		boot->boot.ext.old.CheckSum=-sum;
 		return bootOffset;
 	} else {
 		if(!keepBoot) {
-			boot->jump[0] = 0xeb;
-			boot->jump[1] = 0;
-			boot->jump[2] = 0x90;
-			strncpy(boot->banner, mformat_banner, 8);
+			boot->boot.jump[0] = 0xeb;
+			boot->boot.jump[1] = 0;
+			boot->boot.jump[2] = 0x90;
+			strncpy(boot->boot.banner, mformat_banner, 8);
 			/* It looks like some versions of DOS are
 			 * rather picky about this, and assume default
 			 * parameters without this, ignoring any
@@ -153,7 +155,7 @@ static int init_geometry_boot(struct bootsector *boot, struct device *dev,
 }
 
 
-static int comp_fat_bits(Fs_t *Fs, int estimate, 
+static int comp_fat_bits(Fs_t *Fs, int estimate,
 			 unsigned long tot_sectors, int fat32)
 {
 	int needed_fat_bits;
@@ -192,11 +194,12 @@ static int comp_fat_bits(Fs_t *Fs, int estimate,
 					   Fs->num_fat, 1);
 		if(tot_sectors < min_fat16_size)
 			return 12;
-		else if(tot_sectors >= 2* min_fat16_size)
-			return 16; /* heuristics */
-	}
+ 		else if(Fs->cluster_size == 0 &&
+			tot_sectors >= 2* min_fat16_size)
+ 			return 16; /* heuristics */
+ 	}
 
-	return estimate;
+ 	return estimate;
 }
 
 
@@ -204,16 +207,16 @@ static int comp_fat_bits(Fs_t *Fs, int estimate,
  * According to Microsoft "Hardware White Paper", "Microsoft
  * Extensible Formware Initiative", "FAT32 File System Specification",
  * Version 1.03, December 6, 2000:
- * If (CountofClusters < 4085) { 
- *  // Volume is FAT12  
- * } else if (CountofClusters < 65525) { 
+ * If (CountofClusters < 4085) { // 0x0ff5
+ *  // Volume is FAT12
+ * } else if (CountofClusters < 65525) { // 0xfff5
  *  // Volume is FAT16
- * } else { 
+ * } else {
  *  //Volume is FAT32
  * }
  *
- * This document can be found at the following URL
- * http://www.microsoft.com/hwdev/download/hardware/fatgen103.pdf
+ * This document can be found at the following URL:
+ * https://staff.washington.edu/dittrich/misc/fatgen103.pdf
  * The relevant passus is on page 15.
  *
  * Actually, experimentations with Windows NT 4 show that the
@@ -221,7 +224,7 @@ static int comp_fat_bits(Fs_t *Fs, int estimate,
  * Not sure what the other Microsoft OS'es do though...
  */
 static void calc_fat_bits2(Fs_t *Fs, unsigned long tot_sectors, int fat_bits,
-			   int may_change_cluster_size, 
+			   int may_change_cluster_size,
 			   int may_change_root_size)
 {
 	unsigned long rem_sect;
@@ -239,8 +242,8 @@ static void calc_fat_bits2(Fs_t *Fs, unsigned long tot_sectors, int fat_bits,
 				   Fs->num_fat, Fs->cluster_size)
 
 			if(rem_sect >= MY_DISK_SIZE(16, FAT12+2))
-				/* big enough for FAT16 
-				 * We take a margin of 2, because NT4 
+				/* big enough for FAT16
+				 * We take a margin of 2, because NT4
 				 * misbehaves, and starts considering a disk
 				 * as FAT16 only if it is larger than 4086
 				 * sectors, rather than 4084 as it should
@@ -253,8 +256,8 @@ static void calc_fat_bits2(Fs_t *Fs, unsigned long tot_sectors, int fat_bits,
 				/* "between two chairs",
 				 * augment cluster size, and
 				 * settle it */
-				if(may_change_cluster_size && 
-				   Fs->cluster_size * Fs->sector_size * 2 
+				if(may_change_cluster_size &&
+				   Fs->cluster_size * Fs->sector_size * 2
 				   <= MAX_BYTES_PER_CLUSTER)
 					Fs->cluster_size <<= 1;
 				else if(may_change_root_size) {
@@ -278,7 +281,7 @@ static void calc_fat_bits2(Fs_t *Fs, unsigned long tot_sectors, int fat_bits,
 	}
 }
 
-static __inline__ void format_root(Fs_t *Fs, char *label, struct bootsector *boot)
+static __inline__ void format_root(Fs_t *Fs, char *label, union bootsector *boot)
 {
 	Stream_t *RootDir;
 	char *buf;
@@ -287,7 +290,7 @@ static __inline__ void format_root(Fs_t *Fs, char *label, struct bootsector *boo
 	int dirlen;
 
 	init_clash_handling(&ch);
-	ch.name_converter = label_name;
+	ch.name_converter = label_name_uc;
 	ch.ignore_entry = -2;
 
 	buf = safe_malloc(Fs->sector_size);
@@ -316,13 +319,14 @@ static __inline__ void format_root(Fs_t *Fs, char *label, struct bootsector *boo
 
 	FREE(&RootDir);
 	if(Fs->fat_bits == 32)
-		set_word(boot->dirents, 0);
+		set_word(boot->boot.dirents, 0);
 	else
-		set_word(boot->dirents, Fs->dir_len * (Fs->sector_size / 32));
+		set_word(boot->boot.dirents, Fs->dir_len * (Fs->sector_size / 32));
 	free(buf);
 }
 
 
+#ifdef USE_XDF
 static void xdf_calc_fat_size(Fs_t *Fs, unsigned long tot_sectors,
 			      int fat_bits)
 {
@@ -349,7 +353,7 @@ static void xdf_calc_fat_size(Fs_t *Fs, unsigned long tot_sectors,
 	fprintf(stderr,"Internal error while calculating Xdf fat size\n");
 	exit(1);
 }
-
+#endif
 
 static void calc_fat_size(Fs_t *Fs, unsigned long tot_sectors)
 {
@@ -360,7 +364,7 @@ static void calc_fat_size(Fs_t *Fs, unsigned long tot_sectors)
 	int fat_nybbles;
 	unsigned int slack;
 	int printGrowMsg=1; /* Should we print "growing FAT" messages ?*/
-	
+
 #ifdef DEBUG
 	fprintf(stderr, "Fat start=%d\n", Fs->fat_start);
 	fprintf(stderr, "tot_sectors=%lu\n", tot_sectors);
@@ -382,7 +386,7 @@ static void calc_fat_size(Fs_t *Fs, unsigned long tot_sectors)
 
 	if(Fs->fat_bits == 0) {
 		fprintf(stderr, "Weird, fat bits = 0\n");
-		exit(1);		
+		exit(1);
 	}
 
 
@@ -417,7 +421,7 @@ static void calc_fat_size(Fs_t *Fs, unsigned long tot_sectors)
 		Fs->num_clus = FAT16-1;
 	if(Fs->fat_bits == 12 && Fs->num_clus >= FAT12)
 		Fs->num_clus = FAT12-1;
-	
+
 	/* A safety, if above math is correct, this should not be happen...*/
 	if(Fs->num_clus > (Fs->fat_len * Fs->sector_size * 2 /
 			   fat_nybbles - 2)) {
@@ -495,27 +499,27 @@ static unsigned char bootprog[]=
  0xb9, 0x01, 0x00, 0xcd, 0x13, 0x72, 0x05, 0xea, 0x00, 0x7c, 0x00,
  0x00, 0xcd, 0x19};
 
-static __inline__ void inst_boot_prg(struct bootsector *boot, int offset)
+static __inline__ void inst_boot_prg(union bootsector *boot, int offset)
 {
-	memcpy((char *) boot->jump + offset,
+	memcpy((char *) boot->boot.jump + offset,
 	       (char *) bootprog, sizeof(bootprog) /sizeof(bootprog[0]));
 	if(offset - 2 < 0x80) {
 	  /* short jump */
-	  boot->jump[0] = 0xeb;
-	  boot->jump[1] = offset -2;
-	  boot->jump[2] = 0x90;
+	  boot->boot.jump[0] = 0xeb;
+	  boot->boot.jump[1] = offset -2;
+	  boot->boot.jump[2] = 0x90;
 	} else {
 	  /* long jump, if offset is too large */
-	  boot->jump[0] = 0xe9;
-	  boot->jump[1] = offset -3;
-	  boot->jump[2] = 0x00;
+	  boot->boot.jump[0] = 0xe9;
+	  boot->boot.jump[1] = offset -3;
+	  boot->boot.jump[2] = 0x00;
 	}
-	set_word(boot->jump + offset + 20, offset + 24);
+	set_word(boot->boot.jump + offset + 20, offset + 24);
 }
 
 static void calc_cluster_size(struct Fs_t *Fs, unsigned long tot_sectors,
 			      int fat_bits)
-			
+
 {
 	unsigned int max_clusters; /* maximal possible number of sectors for
 				   * this FAT entry length (12/16/32) */
@@ -525,7 +529,7 @@ static void calc_cluster_size(struct Fs_t *Fs, unsigned long tot_sectors,
 				* the root directory and boot sector(s) */
 
 	switch(abs(fat_bits)) {
-		case 12:			
+		case 12:
 			max_clusters = FAT12-1;
 			max_fat_size = Fs->num_fat *
 				FAT_SIZE(12, Fs->sector_size, max_clusters);
@@ -536,18 +540,37 @@ static void calc_cluster_size(struct Fs_t *Fs, unsigned long tot_sectors,
 			max_fat_size = Fs->num_fat *
 				FAT_SIZE(16, Fs->sector_size, max_clusters);
 			break;
-		case 32:		
-			Fs->cluster_size = 8;
-			/* According to
-			 * http://support.microsoft.com/support/kb/articles/q154/9/97.asp
-			 * Micro$oft does not support FAT32 with less than 4K
+		case 32:
+			/*
+			   FAT32 cluster sizes for disks with 512 block size
+			   according to Microsoft specification fatgen103.doc:
+			
+			   32.5 MB - 260 MB   cluster_size =  1
+			    260 MB -   8 GB   cluster_size =  8
+			      8 GB -  16 GB   cluster_size = 16
+			     16 GB -  32 GB   cluster_size = 32
+			     32 GB -   2 TB   cluster_size = 64
+			
+			   Below calculation is generalized and does not depend
+			   on 512 block size.
 			 */
+			Fs->cluster_size = tot_sectors > 32*1024*1024*2 ? 64 :
+			                   tot_sectors > 16*1024*1024*2 ? 32 :
+			                   tot_sectors >  8*1024*1024*2 ? 16 :
+			                   tot_sectors >     260*1024*2 ? 8 : 1;
 			return;
 		default:
 			fprintf(stderr,"Bad fat size\n");
 			exit(1);
 	}
 
+	if(tot_sectors <= Fs->fat_start + Fs->num_fat + Fs->dir_len) {
+		/* we need at least enough sectors to fit boot, fat and root
+		 * dir */
+		fprintf(stderr, "Not enough sectors\n");
+		exit(1);
+	}
+
 	rem_sect = tot_sectors - Fs->dir_len - Fs->fat_start;
 
 	/* double the cluster size until we can fill up the disk with
@@ -595,7 +618,7 @@ static int old_dos_size_to_geom(size_t size, int *cyls, int *heads, int *sects)
 
 
 static void calc_fs_parameters(struct device *dev, unsigned long tot_sectors,
-			       struct Fs_t *Fs, struct bootsector *boot)
+			       struct Fs_t *Fs, union bootsector *boot)
 {
 	unsigned int i;
 
@@ -607,7 +630,7 @@ static void calc_fs_parameters(struct device *dev, unsigned long tot_sectors,
 		    (Fs->dir_len == 0 || Fs->dir_len == old_dos[i].dir_len) &&
 		    (Fs->cluster_size == 0 ||
 		     Fs->cluster_size == old_dos[i].cluster_size)) {
-			boot->descr = old_dos[i].media;
+			boot->boot.descr = old_dos[i].media;
 			Fs->cluster_size = old_dos[i].cluster_size;
 			Fs->dir_len = old_dos[i].dir_len;
 			Fs->fat_len = old_dos[i].fat_len;
@@ -620,10 +643,10 @@ static void calc_fs_parameters(struct device *dev, unsigned long tot_sectors,
 		int may_change_root_size = (Fs->dir_len == 0);
 
 		/* a non-standard format */
-		if(DWORD(nhs))
-			boot->descr = 0xf8;
+		if(DWORD(nhs) || tot_sectors % (dev->sectors * dev->heads))
+			boot->boot.descr = 0xf8;
 		  else
-			boot->descr = 0xf0;
+			boot->boot.descr = 0xf0;
 
 
 		if(!Fs->cluster_size) {
@@ -635,18 +658,21 @@ static void calc_fs_parameters(struct device *dev, unsigned long tot_sectors,
 					Fs->cluster_size = 1;
 			}
 		}
-		
+
 		if(!Fs->dir_len) {
 			if (dev->heads == 1)
 				Fs->dir_len = 4;
 			else
 				Fs->dir_len = (tot_sectors > 2000) ? 32 : 7;
-		}			
+		}
 
 		calc_cluster_size(Fs, tot_sectors, dev->fat_bits);
+#ifdef USE_XDF
 		if(Fs->fat_len)
 			xdf_calc_fat_size(Fs, tot_sectors, dev->fat_bits);
-		else {
+		else
+#endif
+		{
 			calc_fat_bits2(Fs, tot_sectors, dev->fat_bits,
 				       may_change_cluster_size,
 				       may_change_root_size);
@@ -654,31 +680,32 @@ static void calc_fs_parameters(struct device *dev, unsigned long tot_sectors,
 		}
 	}
 
-	set_word(boot->fatlen, Fs->fat_len);
+	set_word(boot->boot.fatlen, Fs->fat_len);
 }
 
 
 
 static void calc_fs_parameters_32(unsigned long tot_sectors,
-				  struct Fs_t *Fs, struct bootsector *boot)
+				  struct Fs_t *Fs, union bootsector *boot)
 {
 	if(DWORD(nhs))
-		boot->descr = 0xf8;
+		boot->boot.descr = 0xf8;
 	else
-		boot->descr = 0xf0;
+		boot->boot.descr = 0xf0;
+
 	if(!Fs->cluster_size)
-		/* According to
-		 * http://www.microsoft.com/kb/articles/q154/9/97.htm,
-		 * Micro$oft does not support FAT32 with less than 4K
-		 */
-		Fs->cluster_size = 8;
-	
+		calc_cluster_size(Fs, tot_sectors, 32);
 	Fs->dir_len = 0;
 	Fs->num_clus = tot_sectors / Fs->cluster_size;
+	/* Maximal number of clusters on FAT32 is 0xffffff6 */
+	if (Fs->num_clus > 0xffffff6) {
+		fprintf(stderr, "Too many clusters\n");
+		exit(1);
+	}
 	set_fat32(Fs);
 	calc_fat_size(Fs, tot_sectors);
-	set_word(boot->fatlen, 0);
-	set_dword(boot->ext.fat32.bigFat, Fs->fat_len);
+	set_word(boot->boot.fatlen, 0);
+	set_dword(boot->boot.ext.fat32.bigFat, Fs->fat_len);
 }
 
 
@@ -705,9 +732,23 @@ static void usage(int ret)
 }
 
 #ifdef OS_linux
-static int get_block_geom(int fd, struct MT_STAT *buf, struct device *dev,
-			  char *errmsg) {
+static int get_sector_size(int fd, char *errmsg) {
+	int sec_size;
+	if (ioctl(fd, BLKSSZGET, &sec_size) != 0 || sec_size <= 0) {
+		sprintf(errmsg, "Could not get sector size of device (%s)",
+			strerror(errno));
+		return -1;
+	}
+
+	/* Cap sector size at 4096 */
+	if(sec_size > 4096)
+		sec_size = 4096;
+	return sec_size;
+}
+
+static int get_block_geom(int fd, struct device *dev, char *errmsg) {
 	struct hd_geometry geom;
+	int sec_size;
 	long size;
 	int heads=dev->heads;
 	int sectors=dev->sectors;
@@ -718,13 +759,21 @@ static int get_block_geom(int fd, struct MT_STAT *buf, struct device *dev,
 			strerror(errno));
 		return -1;
 	}
-	
+
 	if (ioctl(fd, BLKGETSIZE, &size) < 0) {
 		sprintf(errmsg, "Could not get size of device (%s)",
 			strerror(errno));
 		return -1;
 	}
+
+	sec_size = get_sector_size(fd, errmsg);
+	if(sec_size < 0)
+		return -1;
 	
+	dev->ssize = 0;
+	while (dev->ssize < 0x7F && (128 << dev->ssize) < sec_size)
+		dev->ssize++;
+
 	if(!heads)
 		heads = geom.heads;
 	if(!sectors)
@@ -745,12 +794,97 @@ static int get_block_geom(int fd, struct MT_STAT *buf, struct device *dev,
 	dev->heads = heads;
 	dev->sectors = sectors;
 	if(!dev->tracks)
-		dev->tracks = (size + dev->hidden) / sect_per_track;
-	size = dev->tracks * dev->heads * dev->sectors + dev->hidden;
+		dev->tracks = (size + dev->hidden % sect_per_track) / sect_per_track;
 	return 0;
 }
 #endif
 
+static int get_lba_geom(Stream_t *Direct, unsigned long tot_sectors, struct device *dev, char *errmsg) {
+	int sect_per_track;
+	unsigned long tracks;
+
+	/* if one value is already specified we do not want to overwrite it */
+	if (dev->heads || dev->sectors || dev->tracks) {
+		sprintf(errmsg, "Number of heads or sectors or tracks was already specified");
+		return -1;
+	}
+
+	if (!tot_sectors) {
+#ifdef OS_linux
+		int fd;
+		int sec_size;
+		long size;
+		struct MT_STAT stbuf;
+
+		fd = get_fd(Direct);
+		if (MT_FSTAT(fd, &stbuf) < 0) {
+			sprintf(errmsg, "Could not stat file (%s)", strerror(errno));
+			return -1;
+		}
+
+		if (S_ISBLK(stbuf.st_mode)) {
+			if (ioctl(fd, BLKGETSIZE, &size) != 0) {
+				sprintf(errmsg, "Could not get size of device (%s)",
+					strerror(errno));
+				return -1;
+			}
+			sec_size = get_sector_size(fd, errmsg);
+			if(sec_size < 0)
+				return -1;
+
+			if (!(dev->ssize & 0x80)) {
+				dev->ssize = 0;
+				while (dev->ssize < 0x7F && (128 << dev->ssize) < sec_size)
+				dev->ssize++;
+			}
+			if ((dev->ssize & 0x7f) > 2)
+				tot_sectors = size >> ((dev->ssize & 0x7f) - 2);
+			else
+				tot_sectors = size << (2 - (dev->ssize & 0x7f));
+		} else if (S_ISREG(stbuf.st_mode)) {
+			tot_sectors = stbuf.st_size >> ((dev->ssize & 0x7f) + 7);
+		} else {
+			sprintf(errmsg, "Could not get size of device (%s)",
+				"No method available");
+			return -1;
+		}
+#else
+		mt_size_t size;
+		GET_DATA(Direct, 0, &size, 0, 0);
+		if (size == 0) {
+			sprintf(errmsg, "Could not get size of device (%s)",
+				"No method available");
+			return -1;
+		}
+		tot_sectors = size >> ((dev->ssize & 0x7f) + 7);
+#endif
+	}
+
+	dev->sectors = 63;
+
+	if (tot_sectors < 16*63*1024)
+		dev->heads = 16;
+	else if (tot_sectors < 32*63*1024)
+		dev->heads = 32;
+	else if (tot_sectors < 64*63*1024)
+		dev->heads = 64;
+	else if (tot_sectors < 128*63*1024)
+		dev->heads = 128;
+	else
+		dev->heads = 255;
+
+	sect_per_track = dev->heads * dev->sectors;
+	tracks = (tot_sectors + dev->hidden % sect_per_track) / sect_per_track;
+	if (tracks > 0xFFFFFFFF) {
+		sprintf(errmsg, "Device is too big, it has too many tracks");
+		return -1;
+	}
+
+	dev->tracks = tracks;
+
+	return 0;
+}
+
 void mformat(int argc, char **argv, int dummy)
 {
 	int r; /* generic return value */
@@ -772,18 +906,19 @@ void mformat(int argc, char **argv, int dummy)
 	int format_xdf = 0;
 	struct xdf_info info;
 #endif
-	struct bootsector *boot;
+	union bootsector boot;
 	char *bootSector=0;
 	int c;
 	int keepBoot = 0;
 	struct device used_dev;
 	int argtracks, argheads, argsectors;
-	unsigned long tot_sectors;
+	unsigned long tot_sectors=0;
 	int blocksize;
 
 	char drive, name[EXPAND_BUF];
 
-	char label[VBUFSIZE], buf[MAX_SECTOR];
+	char label[VBUFSIZE];
+
 	dos_name_t shortlabel;
 	struct device *dev;
 	char errmsg[200];
@@ -797,6 +932,10 @@ void mformat(int argc, char **argv, int dummy)
 
 	int Atari = 0; /* should we add an Atari-style serial number ? */
 
+	int backupBoot = 6;
+
+	char *endptr;
+
 	hs = hs_set = 0;
 	argtracks = 0;
 	argheads = 0;
@@ -807,7 +946,7 @@ void mformat(int argc, char **argv, int dummy)
 	serial_set = 0;
 	serial = 0;
 	fsVersion = 0;
-	
+
 	Fs.cluster_size = 0;
 	Fs.refs = 1;
 	Fs.dir_len = 0;
@@ -823,7 +962,7 @@ void mformat(int argc, char **argv, int dummy)
 	  if(Fs.num_fat <= 0)
 	    Fs.num_fat=2;
 	}
-	Fs.Class = &FsClass;	
+	Fs.Class = &FsClass;
 	rate_0 = mtools_rate_0;
 	rate_any = mtools_rate_any;
 
@@ -832,10 +971,11 @@ void mformat(int argc, char **argv, int dummy)
 		usage(0);
 	while ((c = getopt(argc,argv,
 			   "i:148f:t:n:v:qub"
-			   "kB:r:L:IFCc:Xh:s:l:N:H:M:S:2:30:Aad:m:"))!= EOF) {
+			   "kK:B:r:L:I:FCc:Xh:s:T:l:N:H:M:S:2:30:Aad:m:"))!= EOF) {
+		endptr = NULL;
 		switch (c) {
 			case 'i':
-				set_cmd_line_image(optarg, 0);
+				set_cmd_line_image(optarg);
 				break;
 
 			/* standard DOS flags */
@@ -864,6 +1004,10 @@ void mformat(int argc, char **argv, int dummy)
 				argtracks = atoi(optarg);
 				break;
 
+			case 'T':
+				tot_sectors = atoi(optarg);
+				break;
+
 			case 'n': /*non-standard*/
 			case 's':
 				argsectors = atoi(optarg);
@@ -883,7 +1027,7 @@ void mformat(int argc, char **argv, int dummy)
 				fprintf(stderr,
 					"Flag %c not supported by mtools\n",c);
 				exit(1);
-				
+
 
 
 			/* flags added by mtools */
@@ -894,7 +1038,7 @@ void mformat(int argc, char **argv, int dummy)
 
 			case 'S':
 				argssize = atoi(optarg) | 0x80;
-				if(argssize < 0x81)
+				if(argssize < 0x80)
 					usage(1);
 				if(argssize >= 0x87) {
 					fprintf(stderr, "argssize must be less than 6\n");
@@ -935,7 +1079,7 @@ void mformat(int argc, char **argv, int dummy)
 				break;
 
 			case 'N':
- 				serial = strtoul(optarg,0,16);
+ 				serial = strtoul(optarg,&endptr,16);
  				serial_set = 1;
  				break;
 			case 'a': /* Atari style serial number */
@@ -952,7 +1096,7 @@ void mformat(int argc, char **argv, int dummy)
 				break;
 
 			case 'I':
-				fsVersion = strtoul(optarg,0,0);
+				fsVersion = strtoul(optarg,&endptr,0);
 				break;
 
 			case 'c':
@@ -960,10 +1104,10 @@ void mformat(int argc, char **argv, int dummy)
 				break;
 
 			case 'r':
-				Fs.dir_len = strtoul(optarg,0,0);
+				Fs.dir_len = strtoul(optarg,&endptr,0);
 				break;
 			case 'L':
-				Fs.fat_len = strtoul(optarg,0,0);
+				Fs.fat_len = strtoul(optarg,&endptr,0);
 				break;
 
 
@@ -973,6 +1117,13 @@ void mformat(int argc, char **argv, int dummy)
 			case 'k':
 				keepBoot = 1;
 				break;
+			case 'K':
+				backupBoot = atoi(optarg);
+				if(backupBoot < 2 || backupBoot >= 32) {
+				  fprintf(stderr, "Backupboot must be comprised between 2 and 32\n");
+				  exit(1);
+				}
+				break;
 			case 'h':
 				argheads = atoi(optarg);
 				break;
@@ -980,16 +1131,39 @@ void mformat(int argc, char **argv, int dummy)
 				Fs.num_fat = atoi(optarg);
 				break;
 			case 'm':
-				mediaDesc = strtoul(optarg,0,0);
+				mediaDesc = strtoul(optarg,&endptr,0);
+				if(*endptr)
+					mediaDesc = strtoul(optarg,&endptr,16);
 				break;
 			default:
 				usage(1);
 		}
+		if(endptr && *endptr) {
+			fprintf(stderr, "Bad number %s\n", optarg);
+			exit(1);
+		}
 	}
 
-	if (argc - optind != 1 ||
-	    !argv[optind][0] || argv[optind][1] != ':')
+	if (argc - optind > 1)
 		usage(1);
+	if(argc - optind == 1) {
+	    if(!argv[optind][0] || argv[optind][1] != ':')
+		usage(1);
+	    drive = toupper(argv[argc -1][0]);
+	} else {
+	    drive = get_default_drive();
+	    if(drive != ':') {
+	      /* Use default drive only if it is ":" (image file), as else
+		 it would be too dangerous... */
+	      fprintf(stderr, "Drive letter missing\n");
+	      exit(1);
+	    }
+	}
+
+	if(argtracks && tot_sectors) {
+		fprintf(stderr, "Only one of -t or -T may be specified\n");
+		usage(1);
+	}
 
 #ifdef USE_XDF
 	if(create && format_xdf) {
@@ -997,11 +1171,9 @@ void mformat(int argc, char **argv, int dummy)
 		exit(1);
 	}
 #endif
-	
-	drive = toupper(argv[argc -1][0]);
 
-	/* check out a drive whose letter and parameters match */	
-	sprintf(errmsg, "Drive '%c:' not supported", drive);	
+	/* check out a drive whose letter and parameters match */
+	sprintf(errmsg, "Drive '%c:' not supported", drive);
 	Fs.Direct = NULL;
 	blocksize = 0;
 	for(dev=devices;dev->drive;dev++) {
@@ -1018,7 +1190,7 @@ void mformat(int argc, char **argv, int dummy)
 		SET_INT(used_dev.ssize, argssize);
 		if(hs_set)
 			used_dev.hidden = hs;
-		
+
 		expand(dev->name, name);
 #ifdef USING_NEW_VOLD
 		strcpy(name, getVoldName(dev, name));
@@ -1029,14 +1201,11 @@ void mformat(int argc, char **argv, int dummy)
 #endif
 			Fs.Direct = 0;
 #ifdef USE_FLOPPYD
-			Fs.Direct = FloppydOpen(&used_dev, dev, name,
+			Fs.Direct = FloppydOpen(&used_dev, name,
 						O_RDWR | create,
-						errmsg, 0, 1);
-			if(Fs.Direct) {
-				maxSize = max_off_t_31;
-			}
+						errmsg, &maxSize);
 #endif
-			if(!Fs.Direct) {			
+			if(!Fs.Direct) {
 				Fs.Direct = SimpleFileOpen(&used_dev, dev, name,
 							   O_RDWR | create,
 							   errmsg, 0, 1,
@@ -1065,23 +1234,32 @@ void mformat(int argc, char **argv, int dummy)
 
 			if (MT_FSTAT(fd, &stbuf) < 0) {
 				sprintf(errmsg, "Could not stat file (%s)", strerror(errno));
-				continue;						
+				continue;
 			}
 
-			if (S_ISBLK(stbuf.st_mode) &&
-			    get_block_geom(fd, &stbuf, &used_dev, errmsg) < 0)
-				continue;
+			if (S_ISBLK(stbuf.st_mode))
+			    /* If the following get_block_geom fails, do not 
+			     * continue to next drive description, but allow
+			     * get_lba_geom to kick in
+			     */
+			    get_block_geom(fd, &used_dev, errmsg);
 		}
 #endif
 
-		/* no way to find out geometry */
-		if (!used_dev.tracks || !used_dev.heads || !used_dev.sectors){
-			sprintf(errmsg,
-				"Unknown geometry "
-				"(You must tell the complete geometry "
-				"of the disk, \neither in /etc/mtools.conf or "
-				"on the command line) ");
-			continue;
+		if ((!used_dev.tracks && !tot_sectors) ||
+		     !used_dev.heads || !used_dev.sectors){
+			if (get_lba_geom(Fs.Direct, tot_sectors, &used_dev,
+					 errmsg) < 0) {
+				sprintf(errmsg, "%s: "
+					"Complete geometry of the disk "
+					"was not specified, \n"
+					"neither in /etc/mtools.conf nor "
+					"on the command line. \n"
+					"LBA Assist Translation for "
+					"calculating CHS geometry "
+					"of the disk failed.\n", argv[0]);
+				continue;
+			}
 		}
 
 #if 0
@@ -1113,13 +1291,13 @@ void mformat(int argc, char **argv, int dummy)
 			blocksize = Fs.sector_size;
 		else
 			blocksize = used_dev.blocksize;
-		
+
 		if(blocksize > MAX_SECTOR)
 			blocksize = MAX_SECTOR;
 
 		/* do a "test" read */
 		if (!create &&
-		    READS(Fs.Direct, (char *) buf, 0, Fs.sector_size) !=
+		    READS(Fs.Direct, &boot.characters, 0, Fs.sector_size) !=
 		    (signed int) Fs.sector_size) {
 			sprintf(errmsg,
 				"Error reading from '%s', wrong parameters?",
@@ -1130,7 +1308,7 @@ void mformat(int argc, char **argv, int dummy)
 	}
 
 
-	/* print error msg if needed */	
+	/* print error msg if needed */
 	if ( dev->drive == 0 ){
 		FREE(&Fs.Direct);
 		fprintf(stderr,"%s: %s\n", argv[0],errmsg);
@@ -1138,17 +1316,24 @@ void mformat(int argc, char **argv, int dummy)
 	}
 
 	/* calculate the total number of sectors */
-	tot_sectors = used_dev.tracks*used_dev.heads*used_dev.sectors - used_dev.hidden;
+	if(tot_sectors == 0) {
+		unsigned long sect_per_track = used_dev.heads*used_dev.sectors;
+		tot_sectors = used_dev.tracks*sect_per_track - used_dev.hidden%sect_per_track;
+		/* Number of sectors must fit into 32bit value */
+		if (tot_sectors > 0xFFFFFFFF) {
+			fprintf(stderr, "Too few sectors\n");
+			exit(1);
+		}
+	}
 
 	/* create the image file if needed */
 	if (create) {
-		WRITES(Fs.Direct, (char *) buf,
+		WRITES(Fs.Direct, &boot.characters,
 		       sectorsToBytes((Stream_t*)&Fs, tot_sectors-1),
 		       Fs.sector_size);
 	}
 
 	/* the boot sector */
-	boot = (struct bootsector *) buf;
 	if(bootSector) {
 		int fd;
 
@@ -1157,21 +1342,16 @@ void mformat(int argc, char **argv, int dummy)
 			perror("open boot sector");
 			exit(1);
 		}
-		read(fd, buf, blocksize);
-		keepBoot = 1;
-	}
-	if(!keepBoot && !(used_dev.use_2m & 0x7f)) {
-		memset((char *)boot, '\0', Fs.sector_size);
-		if(Fs.sector_size == 512 && !used_dev.partition) {
-			/* install fake partition table pointing to itself */
-			struct partition *partTable=(struct partition *)
-				(((char*) boot) + 0x1ae);
-			setBeginEnd(&partTable[1], 0,
-						used_dev.heads * used_dev.sectors * used_dev.tracks,
-						used_dev.heads, used_dev.sectors, 1, 0);
+		if(read(fd, &boot.bytes, blocksize) < blocksize) {
+			perror("short read on boot sector");
+			exit(1);
 		}
+		keepBoot = 1;
+		close(fd);
 	}
-	set_dword(boot->nhs, used_dev.hidden);
+	if(!keepBoot && !(used_dev.use_2m & 0x7f))
+		memset(boot.characters, '\0', Fs.sector_size);
+	set_dword(boot.boot.nhs, used_dev.hidden);
 
 	Fs.Next = buf_init(Fs.Direct,
 			   blocksize * used_dev.heads * used_dev.sectors,
@@ -1179,48 +1359,61 @@ void mformat(int argc, char **argv, int dummy)
 			   blocksize);
 	Fs.Buffer = 0;
 
-	boot->nfat = Fs.num_fat;
+	boot.boot.nfat = Fs.num_fat;
 	if(!keepBoot)
-		set_word(boot->jump + 510, 0xaa55);
-	
+		set_word(&boot.bytes[510], 0xaa55);
+
 	/* Initialize the remaining parameters */
-	set_word(boot->nsect, used_dev.sectors);
-	set_word(boot->nheads, used_dev.heads);
+	set_word(boot.boot.nsect, used_dev.sectors);
+	set_word(boot.boot.nheads, used_dev.heads);
 
 	used_dev.fat_bits = comp_fat_bits(&Fs,used_dev.fat_bits, tot_sectors, fat32);
 
+	if(!keepBoot && !(used_dev.use_2m & 0x7f)) {
+		if(!used_dev.partition) {
+			/* install fake partition table pointing to itself */
+			struct partition *partTable=(struct partition *)
+				(&boot.bytes[0x1ae]);
+			setBeginEnd(&partTable[1], 0,
+				    used_dev.heads * used_dev.sectors *
+				    used_dev.tracks,
+				    used_dev.heads, used_dev.sectors, 1, 0,
+				    used_dev.fat_bits);
+		}
+	}
+
 	if(used_dev.fat_bits == 32) {
 		Fs.primaryFat = 0;
 		Fs.writeAllFats = 1;
 		Fs.fat_start = 32;
-		calc_fs_parameters_32(tot_sectors, &Fs, boot);
+		calc_fs_parameters_32(tot_sectors, &Fs, &boot);
 
 		Fs.clus_start = Fs.num_fat * Fs.fat_len + Fs.fat_start;
 
 		/* extension flags: mirror fats, and use #0 as primary */
-		set_word(boot->ext.fat32.extFlags,0);
+		set_word(boot.boot.ext.fat32.extFlags,0);
 
 		/* fs version.  What should go here? */
-		set_word(boot->ext.fat32.fsVersion,fsVersion);
+		set_word(boot.boot.ext.fat32.fsVersion,fsVersion);
 
 		/* root directory */
-		set_dword(boot->ext.fat32.rootCluster, Fs.rootCluster = 2);
+		set_dword(boot.boot.ext.fat32.rootCluster, Fs.rootCluster = 2);
 
 		/* info sector */
-		set_word(boot->ext.fat32.infoSector, Fs.infoSectorLoc = 1);
+		set_word(boot.boot.ext.fat32.infoSector, Fs.infoSectorLoc = 1);
 		Fs.infoSectorLoc = 1;
 
 		/* no backup boot sector */
-		set_word(boot->ext.fat32.backupBoot, 6);
-		
-		labelBlock = & boot->ext.fat32.labelBlock;
+		set_word(boot.boot.ext.fat32.backupBoot, backupBoot);
+
+		labelBlock = & boot.boot.ext.fat32.labelBlock;
 	} else {
 		Fs.infoSectorLoc = 0;
 		Fs.fat_start = 1;
-		calc_fs_parameters(&used_dev, tot_sectors, &Fs, boot);
+		calc_fs_parameters(&used_dev, tot_sectors, &Fs, &boot);
 		Fs.dir_start = Fs.num_fat * Fs.fat_len + Fs.fat_start;
 		Fs.clus_start = Fs.dir_start + Fs.dir_len;
-		labelBlock = & boot->ext.old.labelBlock;
+		labelBlock = & boot.boot.ext.old.labelBlock;
 
 	}
 
@@ -1240,48 +1433,49 @@ void mformat(int argc, char **argv, int dummy)
 		srandom((long)time (0));
 	if (!serial_set)
 		serial=random();
-	set_dword(labelBlock->serial, serial);	
-	label_name(GET_DOSCONVERT((Stream_t *)&Fs),
-		   label[0] ? label : "NO NAME    ", 0, &mangled, &shortlabel);
+	set_dword(labelBlock->serial, serial);
+	label_name_pc(GET_DOSCONVERT((Stream_t *)&Fs),
+		      label[0] ? label : "NO NAME    ", 0,
+		      &mangled, &shortlabel);
 	strncpy(labelBlock->label, shortlabel.base, 11);
 	sprintf(labelBlock->fat_type, "FAT%2.2d  ", Fs.fat_bits);
 	labelBlock->fat_type[7] = ' ';
 
-	set_word(boot->secsiz, Fs.sector_size);
-	boot->clsiz = (unsigned char) Fs.cluster_size;
-	set_word(boot->nrsvsect, Fs.fat_start);
+	set_word(boot.boot.secsiz, Fs.sector_size);
+	boot.boot.clsiz = (unsigned char) Fs.cluster_size;
+	set_word(boot.boot.nrsvsect, Fs.fat_start);
 
-	bootOffset = init_geometry_boot(boot, &used_dev, sectors0,
+	bootOffset = init_geometry_boot(&boot, &used_dev, sectors0,
 					rate_0, rate_any,
 					&tot_sectors, keepBoot);
 	if(!bootOffset) {
-		bootOffset = ((char *) labelBlock) - ((char *) boot) +
+		bootOffset = ((unsigned char *) labelBlock) - boot.bytes +
 			sizeof(struct label_blk_t);
 	}
 	if(Atari) {
-		boot->banner[4] = 0;
-		boot->banner[5] = random();
-		boot->banner[6] = random();
-		boot->banner[7] = random();
-	}		
+		boot.boot.banner[4] = 0;
+		boot.boot.banner[5] = random();
+		boot.boot.banner[6] = random();
+		boot.boot.banner[7] = random();
+	}
 
 	if(!keepBoot)
-		inst_boot_prg(boot, bootOffset);
+		inst_boot_prg(&boot, bootOffset);
 	/* Mimic 3.8 behavior, else 2m disk do not work (???)
 	 * luferbu@fluidsignal.com (Luis Bustamante), Fri, 14 Jun 2002
 	 */
 	if(used_dev.use_2m & 0x7f) {
-	  boot->jump[0] = 0xeb;
-	  boot->jump[1] = 0x80;
-	  boot->jump[2] = 0x90;
+	  boot.boot.jump[0] = 0xeb;
+	  boot.boot.jump[1] = 0x80;
+	  boot.boot.jump[2] = 0x90;
 	}
 	if(used_dev.use_2m & 0x7f)
 		Fs.num_fat = 1;
 	if(mediaDesc != -1)
-		boot->descr=mediaDesc;
+		boot.boot.descr=mediaDesc;
 	Fs.lastFatSectorNr = 0;
 	Fs.lastFatSectorData = 0;
-	zero_fat(&Fs, boot->descr);
+	zero_fat(&Fs, boot.boot.descr);
 	Fs.freeSpace = Fs.num_clus;
 	Fs.last = 2;
 
@@ -1293,11 +1487,19 @@ void mformat(int argc, char **argv, int dummy)
 			fatEncode(&Fs, i+2, 0xfff7);
 #endif
 
-	format_root(&Fs, label, boot);
-	WRITES((Stream_t *)&Fs, (char *) boot, (mt_off_t) 0, Fs.sector_size);
-	if(Fs.fat_bits == 32 && WORD(ext.fat32.backupBoot) != MAX16) {
-		WRITES((Stream_t *)&Fs, (char *) boot,
-		       sectorsToBytes((Stream_t*)&Fs, WORD(ext.fat32.backupBoot)),
+	format_root(&Fs, label, &boot);
+	WRITES((Stream_t *)&Fs, boot.characters,
+	       (mt_off_t) 0, Fs.sector_size);
+
+	if(used_dev.fat_bits == 32) {
+	  WRITES((Stream_t *)&Fs, boot.characters,
+		 (mt_off_t) backupBoot * Fs.sector_size, Fs.sector_size);
+	}
+
+	if(Fs.fat_bits == 32 && WORD_S(ext.fat32.backupBoot) != MAX16) {
+		WRITES((Stream_t *)&Fs, boot.characters,
+		       sectorsToBytes((Stream_t*)&Fs,
+				      WORD_S(ext.fat32.backupBoot)),
 		       Fs.sector_size);
 	}
 	FLUSH((Stream_t *)&Fs); /* flushes Fs.
@@ -1313,7 +1515,7 @@ void mformat(int argc, char **argv, int dummy)
 			"Bourne shell syntax (sh, ash, bash, ksh, zsh etc):\n"
 			" export MTOOLS_USE_XDF=1\n\n"
 			"C shell syntax (csh and tcsh):\n"
-			" setenv MTOOLS_USE_XDF 1\n" );	
+			" setenv MTOOLS_USE_XDF 1\n" );
 #endif
 	exit(0);
 }
