@@ -1,7 +1,23 @@
 #ifndef MTOOLS_MSDOS_H
 #define MTOOLS_MSDOS_H
 
-/*
+/*  Copyright 1986-1992 Emmet P. Gray.
+ *  Copyright 1996-1998,2000-2003,2006,2007,2009 Alain Knaff.
+ *  This file is part of mtools.
+ *
+ *  Mtools is free software: you can redistribute it and/or modify
+ *  it under the terms of the GNU General Public License as published by
+ *  the Free Software Foundation, either version 3 of the License, or
+ *  (at your option) any later version.
+ *
+ *  Mtools is distributed in the hope that it will be useful,
+ *  but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *  GNU General Public License for more details.
+ *
+ *  You should have received a copy of the GNU General Public License
+ *  along with Mtools.  If not, see <http://www.gnu.org/licenses/>.
+ *
  * msdos common header file
  */
 
@@ -21,6 +37,7 @@
 #define _DWORD(x) ((unsigned int)(_WORD(x) + (_WORD((x)+2) << 16)))
 
 #define DELMARK ((char) 0xe5)
+#define ENDMARK ((char) 0x00)
 
 struct directory {
 	char name[8];			/*  0 file name */
@@ -143,15 +160,15 @@ typedef struct oldboot_t {
 	unsigned char junk[1024 - 80];	/* 80 remaining data */
 } oldboot_t;
 
-struct bootsector {
+struct bootsector_s {
 	unsigned char jump[3];		/* 0  Jump to boot code */
-	char banner[8] PACKED;	       	/* 3  OEM name & version */
-	unsigned char secsiz[2] PACKED;	/* 11 Bytes per sector hopefully 512 */
+	char banner[8];	       		/* 3  OEM name & version */
+	unsigned char secsiz[2];	/* 11 Bytes per sector hopefully 512 */
 	unsigned char clsiz;    	/* 13 Cluster size in sectors */
 	unsigned char nrsvsect[2];	/* 14 Number of reserved (boot) sectors */
 	unsigned char nfat;		/* 16 Number of FAT tables hopefully 2 */
-	unsigned char dirents[2] PACKED;/* 17 Number of directory slots */
-	unsigned char psect[2] PACKED; 	/* 19 Total sectors on disk */
+	unsigned char dirents[2];	/* 17 Number of directory slots */
+	unsigned char psect[2]; 	/* 19 Total sectors on disk */
 	unsigned char descr;		/* 21 Media descriptor=first byte of FAT */
 	unsigned char fatlen[2];	/* 22 Sectors in FAT */
 	unsigned char nsect[2];		/* 24 Sectors/track */
@@ -164,12 +181,22 @@ struct bootsector {
 		struct oldboot_t old;
 	} ext;
 };
+
 #define MAX_BOOT 4096
 
+union bootsector {
+	unsigned char bytes[MAX_BOOT];
+	char characters[MAX_BOOT];
+	struct bootsector_s boot;
+};
 
 #define CHAR(x) (boot->x[0])
-#define WORD(x) (_WORD(boot->x))
-#define DWORD(x) (_DWORD(boot->x))
+#define WORD(x) (_WORD(boot->boot.x))
+#define DWORD(x) (_DWORD(boot->boot.x))
+
+#define WORD_S(x) (_WORD(boot.boot.x))
+#define DWORD_S(x) (_DWORD(boot.boot.x))
+
 #define OFFSET(x) (((char *) (boot->x)) - ((char *)(boot->jump)))
 
 
@@ -187,15 +214,16 @@ extern struct OldDos_t {
 
 /* max FAT12/FAT16 sizes, according to
    
- http://www.microsoft.com/hwdev/download/hardware/fatgen103.pdf
+ https://staff.washington.edu/dittrich/misc/fatgen103.pdf
+ https://download.microsoft.com/download/1/6/1/161ba512-40e2-4cc9-843a-923143f3456c/fatgen103.doc
 
  interestingly enough, another Microsoft document 
  [http://support.microsoft.com/default.aspx?scid=kb%3ben-us%3b67321]
  gives different values, but the first seems to be more sure about
  itself, so we believe that one ;-)
 */
-#define FAT12 4085 /* max. number of clusters described by a 12 bit FAT */
-#define FAT16 65525 /* max number of clusters for a 16 bit FAT */
+#define FAT12 0x0ff5 /* max. number + 1 of clusters described by a 12 bit FAT */
+#define FAT16 0xfff5 /* max number + 1 of clusters for a 16 bit FAT */
 
 #define ATTR_ARCHIVE 0x20
 #define ATTR_DIR 0x10
@@ -237,6 +265,7 @@ extern struct OldDos_t {
 
 extern const char *mversion;
 extern const char *mdate;
+extern const char *mformat_banner;
 
 extern char *Version;
 extern char *Date;
